# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import asdict
from typing import Any, Callable, Sequence

from dkg.exceptions import ValidationError
from dkg.managers.async_manager import AsyncRequestManager
from dkg.method import Method
from dkg.types import TReturn


class AsyncModule:
    manager: AsyncRequestManager

    def retrieve_caller_fn(
        self, method: Method[Callable[..., TReturn]]
    ) -> Callable[..., TReturn]:
        async def caller(*args: Any, **kwargs: Any) -> TReturn:
            processed_args = method.process_args(*args, **kwargs)
            request_params = asdict(method.action)
            request_params.update(processed_args)

            return await self.manager.blocking_request(
                type(method.action), request_params
            )

        return caller

    def _attach_modules(self, module_definitions: dict[str, Any]) -> None:
        for module_name, module_info in module_definitions.items():
            module_info_is_list_like = isinstance(module_info, Sequence)

            module = module_info[0] if module_info_is_list_like else module_info

            if hasattr(self, module_name):
                raise AttributeError(
                    f"Cannot set {self} module named '{module_name}'. "
                    " The dkg object already has an attribute with that name"
                )

            setattr(self, module_name, module)

            if module_info_is_list_like:
                if len(module_info) == 2:
                    submodule_definitions = module_info[1]
                    module: "AsyncModule" = getattr(self, module_name)
                    module._attach_modules(submodule_definitions)
                elif len(module_info) != 1:
                    raise ValidationError(
                        "Module definitions can only have 1 or 2 elements."
                    )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Any, Type

from dkg.dataclasses import BlockchainResponseDict, NodeResponseDict
from dkg.exceptions import InvalidRequest
from dkg.providers import BlockchainProvider, NodeHTTPProvider
from dkg.utils.blockchain_request import ContractInteraction, JSONRPCRequest
from dkg.utils.node_request import NodeCall


class DefaultRequestManager:
    def __init__(
        self, node_provider: NodeHTTPProvider, blockchain_provider: BlockchainProvider
    ):
        self.node_provider = node_provider
        self.blockchain_provider = blockchain_provider

    @property
    def node_provider(self) -> NodeHTTPProvider:
        return self._node_provider

    @node_provider.setter
    def node_provider(self, node_provider: NodeHTTPProvider) -> None:
        self._node_provider = node_provider

    @property
    def blockchain_provider(
        self,
    ) -> BlockchainProvider:
        return self._blockchain_provider

    @blockchain_provider.setter
    def blockchain_provider(self, blockchain_provider: BlockchainProvider) -> None:
        self._blockchain_provider = blockchain_provider

    def blocking_request(
        self,
        request_type: Type[JSONRPCRequest | ContractInteraction | NodeCall],
        request_params: dict[str, Any],
    ) -> BlockchainResponseDict | NodeResponseDict:
        if issubclass(request_type, JSONRPCRequest):
            return self.blockchain_provider.make_json_rpc_request(**request_params)
        elif issubclass(request_type, ContractInteraction):
            return self.blockchain_provider.call_function(**request_params)
        elif issubclass(request_type, NodeCall):
            return self.node_provider.make_request(**request_params)
        else:
            raise InvalidRequest(
                "Invalid Request. Manager can only process Blockchain/Node requests."
            )



from .blockchain import BlockchainProvider, AsyncBlockchainProvider
from .node import NodeHTTPProvider, AsyncNodeHTTPProvider

__all__ = [
    "BlockchainProvider",
    "AsyncBlockchainProvider",
    "NodeHTTPProvider",
    "AsyncNodeHTTPProvider",
]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import json
import asyncio
from typing import Literal
from pyld import jsonld
from web3 import Web3
from web3.constants import ADDRESS_ZERO
from web3.types import TxReceipt
from itertools import chain
from eth_account.messages import encode_defunct
from eth_account import Account
from hexbytes import HexBytes

from dkg.constants import (
    PRIVATE_ASSERTION_PREDICATE,
    PRIVATE_RESOURCE_PREDICATE,
    PRIVATE_HASH_SUBJECT_PREFIX,
    CHUNK_BYTE_SIZE,
    MAX_FILE_SIZE,
    DEFAULT_RDF_FORMAT,
    Operations,
    OutputTypes,
)
from dkg.managers.async_manager import AsyncRequestManager
from dkg.method import Method
from dkg.modules.async_module import AsyncModule
from dkg.types import JSONLD, UAL, Address, HexStr
from dkg.utils.blockchain_request import (
    BlockchainRequest,
)
from dkg.utils.node_request import (
    OperationStatus,
)
from dkg.utils.ual import format_ual, parse_ual
import dkg.utils.knowledge_collection_tools as kc_tools
import dkg.utils.knowledge_asset_tools as ka_tools
from dkg.services.input_service import InputService
from dkg.services.node_services.async_node_service import AsyncNodeService
from dkg.services.blockchain_services.async_blockchain_service import (
    AsyncBlockchainService,
)
from dkg.utils.node_request import get_operation_status_object


class AsyncKnowledgeAsset(AsyncModule):
    def __init__(
        self,
        manager: AsyncRequestManager,
        input_service: InputService,
        node_service: AsyncNodeService,
        blockchain_service: AsyncBlockchainService,
    ):
        self.manager = manager
        self.input_service = input_service
        self.node_service = node_service
        self.blockchain_service = blockchain_service

    async def is_valid_ual(self, ual: UAL) -> bool:
        if not ual or not isinstance(ual, str):
            raise ValueError("UAL must be a non-empty string.")

        parts = ual.split("/")
        if len(parts) != 3:
            raise ValueError("UAL format is incorrect.")

        prefixes = parts[0].split(":")
        prefixes_number = len(prefixes)
        if prefixes_number != 3 and prefixes_number != 4:
            raise ValueError("Prefix format in UAL is incorrect.")

        if prefixes[0] != "did":
            raise ValueError(
                f"Invalid DID prefix. Expected: 'did'. Received: '{prefixes[0]}'."
            )

        if prefixes[1] != "dkg":
            raise ValueError(
                f"Invalid DKG prefix. Expected: 'dkg'. Received: '{prefixes[1]}'."
            )

        if prefixes[2] != (
            blockchain_name := (
                self.manager.blockchain_provider.blockchain_id.split(":")[0]
            )
        ):
            raise ValueError(
                "Invalid blockchain name in the UAL prefix. "
                f"Expected: '{blockchain_name}'. Received: '${prefixes[2]}'."
            )

        if prefixes_number == 4:
            chain_id = self.manager.blockchain_provider.blockchain_id.split(":")[1]

            if int(prefixes[3]) != int(chain_id):
                raise ValueError(
                    "Chain ID in UAL does not match the blockchain. "
                    f"Expected: '${chain_id}'. Received: '${prefixes[3]}'."
                )

        contract_address = self.manager.blockchain_provider.contracts[
            "ContentAssetStorage"
        ].address

        if parts[1].lower() != contract_address.lower():
            raise ValueError(
                "Contract address in UAL does not match. "
                f"Expected: '${contract_address.lower()}'. "
                f"Received: '${parts[1].lower()}'."
            )

        try:
            owner = await self.blockchain_service.get_owner(int(parts[2]))

            if not owner or owner == ADDRESS_ZERO:
                raise ValueError("Token does not exist or has no owner.")

            return True
        except Exception as err:
            raise ValueError(f"Error fetching asset owner: {err}")

    def process_content(self, content: str) -> list:
        return [line.strip() for line in content.split("\n") if line.strip() != ""]

    def insert_triple_sorted(self, triples_list: list, new_triple: str) -> int:
        # Assuming triples_list is already sorted
        left = 0
        right = len(triples_list)

        while left < right:
            mid = (left + right) // 2
            if triples_list[mid] < new_triple:
                left = mid + 1
            else:
                right = mid

        # Insert the new triple at the correct position
        triples_list.insert(left, new_triple)
        return left

    def get_operation_status_dict(self, operation_result, operation_id):
        # Check if data exists and has errorType
        operation_data = (
            {"status": operation_result.get("status"), **operation_result.get("data")}
            if operation_result.get("data")
            and operation_result.get("data", {}).get("errorType")
            else {"status": operation_result.get("status")}
        )

        return {"operationId": operation_id, **operation_data}

    def get_message_signer_address(self, dataset_root: str, signature: dict):
        message = encode_defunct(HexBytes(dataset_root))
        r, s, v = signature.get("r"), signature.get("s"), signature.get("v")
        r = r[2:] if r.startswith("0x") else r
        s = s[2:] if s.startswith("0x") else s

        sig = "0x" + r + s + hex(v)[2:].zfill(2)

        return Account.recover_message(message, signature=sig)

    async def process_signatures(self, signatures, dataset_root):
        async def process_signature(signature):
            try:
                signer_address = self.get_message_signer_address(
                    dataset_root, signature
                )

                key_is_operational_wallet = (
                    await self.blockchain_service.key_is_operational_wallet(
                        signature.get("identityId"),
                        Web3.solidity_keccak(["address"], [signer_address]),
                        2,  # IdentityLib.OPERATIONAL_KEY
                    )
                )

                if key_is_operational_wallet:
                    return {
                        "identityId": signature.get("identityId"),
                        "r": signature.get("r"),
                        "vs": signature.get("vs"),
                    }

            except Exception:
                pass
            return None

        # Run all signature processing concurrently
        results = await asyncio.gather(
            *(process_signature(sig) for sig in signatures), return_exceptions=False
        )

        # Filter out None results and organize the data
        valid_results = [r for r in results if r is not None]

        return {
            "identity_ids": [r["identityId"] for r in valid_results],
            "r": [r["r"] for r in valid_results],
            "vs": [r["vs"] for r in valid_results],
        }

    async def create(
        self,
        content: dict[Literal["public", "private"], JSONLD],
        options: dict = None,
    ) -> dict[str, UAL | HexStr | dict[str, dict[str, str] | TxReceipt]]:
        if options is None:
            options = {}

        arguments = self.input_service.get_asset_create_arguments(options)

        max_number_of_retries = arguments.get("max_number_of_retries")
        frequency = arguments.get("frequency")
        epochs_num = arguments.get("epochs_num")
        hash_function_id = arguments.get("hash_function_id")
        immutable = arguments.get("immutable")
        token_amount = arguments.get("token_amount")
        payer = arguments.get("payer")
        minimum_number_of_finalization_confirmations = arguments.get(
            "minimum_number_of_finalization_confirmations"
        )
        minimum_number_of_node_replications = arguments.get(
            "minimum_number_of_node_replications"
        )
        blockchain_id = self.manager.blockchain_provider.blockchain_id

        dataset = {}
        public_content = dataset.get("public")
        private_content = dataset.get("private")
        if isinstance(content, str):
            dataset["public"] = self.process_content(content)
        elif isinstance(public_content, str) or (
            not public_content and private_content and isinstance(private_content, str)
        ):
            if public_content:
                dataset["public"] = self.process_content(public_content)
            else:
                dataset["public"] = []

            if private_content and isinstance(private_content, str):
                dataset["private"] = self.process_content(private_content)
        else:
            dataset = kc_tools.format_dataset(content)

        public_triples_grouped = []

        dataset["public"] = kc_tools.generate_missing_ids_for_blank_nodes(
            dataset.get("public")
        )

        if dataset.get("private") and len(dataset.get("private")):
            dataset["private"] = kc_tools.generate_missing_ids_for_blank_nodes(
                dataset.get("private")
            )

            # Group private triples by subject and flatten
            private_triples_grouped = kc_tools.group_nquads_by_subject(
                dataset.get("private"), True
            )

            dataset["private"] = list(chain.from_iterable(private_triples_grouped))

            # Compute private root and add to public
            private_root = kc_tools.calculate_merkle_root(dataset.get("private"))
            dataset["public"].append(
                f'<{ka_tools.generate_named_node()}> <{PRIVATE_ASSERTION_PREDICATE}> "{private_root}" .'
            )

            # Compute private root and add to public
            public_triples_grouped = kc_tools.group_nquads_by_subject(
                dataset.get("public"), True
            )

            # Create a dictionary for public subject -> index for quick lookup
            public_subject_dict = {}
            for i in range(len(public_triples_grouped)):
                public_subject = public_triples_grouped[i][0].split(" ")[0]
                public_subject_dict[public_subject] = i

            private_triple_subject_hashes_grouped_without_public_pair = []

            # Integrate private subjects into public or store separately if no match to be appended later
            for private_triples in private_triples_grouped:
                private_subject = private_triples[0].split(" ")[
                    0
                ]  # Extract the private subject

                private_subject_hash = kc_tools.solidity_packed_sha256(
                    types=["string"],
                    values=[private_subject[1:-1]],
                )

                if (
                    private_subject in public_subject_dict
                ):  # Check if there's a public pair
                    # If there's a public pair, insert a representation in that group
                    public_index = public_subject_dict.get(private_subject)
                    self.insert_triple_sorted(
                        public_triples_grouped[public_index],
                        f"{private_subject} <{PRIVATE_RESOURCE_PREDICATE}> <{ka_tools.generate_named_node()}> .",
                    )
                else:
                    # If no public pair, maintain separate list, inserting sorted by hash
                    self.insert_triple_sorted(
                        private_triple_subject_hashes_grouped_without_public_pair,
                        f"<{PRIVATE_HASH_SUBJECT_PREFIX}{private_subject_hash}> <{PRIVATE_RESOURCE_PREDICATE}> <{ka_tools.generate_named_node()}> .",
                    )

            for triple in private_triple_subject_hashes_grouped_without_public_pair:
                public_triples_grouped.append([triple])

            dataset["public"] = list(chain.from_iterable(public_triples_grouped))
        else:
            # No private triples, just group and flatten public
            public_triples_grouped = kc_tools.group_nquads_by_subject(
                dataset.get("public"), True
            )
            dataset["public"] = list(chain.from_iterable(public_triples_grouped))

        # Calculate the number of chunks
        number_of_chunks = kc_tools.calculate_number_of_chunks(
            dataset.get("public"), CHUNK_BYTE_SIZE
        )
        dataset_size = number_of_chunks * CHUNK_BYTE_SIZE

        # Validate the assertion size in bytes
        if dataset_size > MAX_FILE_SIZE:
            raise ValueError(f"File size limit is {MAX_FILE_SIZE / (1024 * 1024)}MB.")

        # Calculate the Merkle root
        dataset_root = kc_tools.calculate_merkle_root(dataset.get("public"))

        # Get the contract address for KnowledgeCollectionStorage
        content_asset_storage_address = (
            await self.blockchain_service.get_asset_storage_address(
                "KnowledgeCollectionStorage"
            )
        )

        result = await self.node_service.publish(
            dataset_root,
            dataset,
            blockchain_id,
            hash_function_id,
            minimum_number_of_node_replications,
        )
        publish_operation_id = result.get("operationId")
        publish_operation_result = await self.node_service.get_operation_result(
            publish_operation_id,
            Operations.PUBLISH.value,
            max_number_of_retries,
            frequency,
        )

        if publish_operation_result.get(
            "status"
        ) != OperationStatus.COMPLETED and not publish_operation_result.get(
            "data", {}
        ).get("minAcksReached"):
            return {
                "datasetRoot": dataset_root,
                "operation": {
                    "publish": self.get_operation_status_dict(
                        publish_operation_result, publish_operation_id
                    )
                },
            }

        data = publish_operation_result.get("data", {})
        signatures = data.get("signatures")

        publisher_node_signature = data.get("publisherNodeSignature", {})
        publisher_node_identity_id = publisher_node_signature.get("identityId")
        publisher_node_r = publisher_node_signature.get("r")
        publisher_node_vs = publisher_node_signature.get("vs")

        results = await self.process_signatures(signatures, dataset_root)
        identity_ids = results["identity_ids"]
        r = results["r"]
        vs = results["vs"]

        if token_amount:
            estimated_publishing_cost = token_amount
        else:
            time_until_next_epoch = (
                await self.blockchain_service.time_until_next_epoch()
            )
            epoch_length = await self.blockchain_service.epoch_length()
            stake_weighted_average_ask = (
                await self.blockchain_service.get_stake_weighted_average_ask()
            )

            # Convert to integers and perform calculation
            estimated_publishing_cost = (
                (
                    int(stake_weighted_average_ask)
                    * (
                        int(epochs_num) * int(1e18)
                        + (int(time_until_next_epoch) * int(1e18)) // int(epoch_length)
                    )
                    * int(dataset_size)
                )
                // 1024
                // int(1e18)
            )

        knowledge_collection_id = None
        mint_knowledge_asset_receipt = None

        knowledge_collection_result = (
            await self.blockchain_service.create_knowledge_collection(
                {
                    "publishOperationId": publish_operation_id,
                    "merkleRoot": dataset_root,
                    "knowledgeAssetsAmount": kc_tools.count_distinct_subjects(
                        dataset.get("public")
                    ),
                    "byteSize": dataset_size,
                    "epochs": epochs_num,
                    "tokenAmount": estimated_publishing_cost,
                    "isImmutable": immutable,
                    "paymaster": payer,
                    "publisherNodeIdentityId": publisher_node_identity_id,
                    "publisherNodeR": publisher_node_r,
                    "publisherNodeVS": publisher_node_vs,
                    "identityIds": identity_ids,
                    "r": r,
                    "vs": vs,
                },
                None,
                None,
            )
        )
        knowledge_collection_id = knowledge_collection_result.knowledge_collection_id
        mint_knowledge_asset_receipt = knowledge_collection_result.receipt

        ual = format_ual(
            blockchain_id, content_asset_storage_address, knowledge_collection_id
        )

        finality_status_result = 0
        if minimum_number_of_finalization_confirmations > 0:
            finality_status_result = await self.node_service.finality_status(
                ual,
                minimum_number_of_finalization_confirmations,
                max_number_of_retries,
                frequency,
            )

        return json.loads(
            Web3.to_json(
                {
                    "UAL": ual,
                    "datasetRoot": dataset_root,
                    "signatures": publish_operation_result.get("data", {}).get(
                        "signatures"
                    ),
                    "operation": {
                        "mintKnowledgeAsset": mint_knowledge_asset_receipt,
                        "publish": get_operation_status_object(
                            publish_operation_result, publish_operation_id
                        ),
                        "finality": {
                            "status": "FINALIZED"
                            if finality_status_result
                            >= minimum_number_of_finalization_confirmations
                            else "NOT FINALIZED"
                        },
                        "numberOfConfirmations": finality_status_result,
                        "requiredConfirmations": minimum_number_of_finalization_confirmations,
                    },
                }
            )
        )

    _submit_knowledge_asset = Method(BlockchainRequest.submit_knowledge_asset)

    def submit_to_paranet(
        self, ual: UAL, paranet_ual: UAL
    ) -> dict[str, UAL | Address | TxReceipt]:
        parsed_ual = parse_ual(ual)
        knowledge_asset_storage, knowledge_asset_token_id = (
            parsed_ual["contract_address"],
            parsed_ual["token_id"],
        )

        parsed_paranet_ual = parse_ual(paranet_ual)
        paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id = (
            parsed_paranet_ual["contract_address"],
            parsed_paranet_ual["token_id"],
        )

        receipt: TxReceipt = self._submit_knowledge_asset(
            paranet_knowledge_asset_storage,
            paranet_knowledge_asset_token_id,
            knowledge_asset_storage,
            knowledge_asset_token_id,
        )

        return {
            "UAL": ual,
            "paranetUAL": paranet_ual,
            "paranetId": Web3.to_hex(
                Web3.solidity_keccak(
                    ["address", "uint256"],
                    [knowledge_asset_storage, knowledge_asset_token_id],
                )
            ),
            "operation": json.loads(Web3.to_json(receipt)),
        }

    _transfer = Method(BlockchainRequest.transfer_asset)

    def transfer(
        self,
        ual: UAL,
        new_owner: Address,
    ) -> dict[str, UAL | Address | TxReceipt]:
        token_id = parse_ual(ual)["token_id"]

        receipt: TxReceipt = self._transfer(
            self.manager.blockchain_provider.account,
            new_owner,
            token_id,
        )

        return {
            "UAL": ual,
            "owner": new_owner,
            "operation": json.loads(Web3.to_json(receipt)),
        }

    _burn_asset = Method(BlockchainRequest.burn_asset)

    def burn(self, ual: UAL) -> dict[str, UAL | TxReceipt]:
        token_id = parse_ual(ual)["token_id"]

        receipt: TxReceipt = self._burn_asset(token_id)

        return {"UAL": ual, "operation": json.loads(Web3.to_json(receipt))}

    _get_latest_assertion_id = Method(BlockchainRequest.get_latest_assertion_id)

    async def get(self, ual: UAL, options: dict = None) -> dict:
        if options is None:
            options = {}

        arguments = self.input_service.get_asset_get_arguments(options)

        max_number_of_retries = arguments.get("max_number_of_retries")
        frequency = arguments.get("frequency")
        state = arguments.get("state")
        include_metadata = arguments.get("include_metadata")
        content_type = arguments.get("content_type")
        validate = arguments.get("validate")
        output_format = arguments.get("output_format")
        hash_function_id = arguments.get("hash_function_id")
        paranet_ual = arguments.get("paranet_ual")
        subject_ual = arguments.get("subject_ual")

        ual_with_state = f"{ual}:{state}" if state else ual
        result = await self.node_service.get(
            ual_with_state,
            content_type,
            include_metadata,
            hash_function_id,
            paranet_ual,
            subject_ual,
        )
        get_public_operation_id = result.get("operationId")
        get_public_operation_result = await self.node_service.get_operation_result(
            get_public_operation_id,
            Operations.GET.value,
            max_number_of_retries,
            frequency,
        )

        if subject_ual:
            if get_public_operation_result.get("data"):
                return {
                    "operation": {
                        "get": get_operation_status_object(
                            get_public_operation_result, get_public_operation_id
                        ),
                    },
                    "subject_ual_pairs": get_public_operation_result.get("data"),
                }
            if get_public_operation_result.get("status") != "FAILED":
                get_public_operation_result["data"] = {
                    "errorType": "DKG_CLIENT_ERROR",
                    "errorMessage": "Unable to find assertion on the network!",
                }
                get_public_operation_result["status"] = "FAILED"

            return {
                "operation": {
                    "get": get_operation_status_object(
                        get_public_operation_result, get_public_operation_id
                    ),
                },
            }
        metadata = get_public_operation_result.get("data")
        assertion = get_public_operation_result.get("data", {}).get("assertion", None)

        if not assertion:
            if get_public_operation_result.get("status") != "FAILED":
                get_public_operation_result["data"] = {
                    "errorType": "DKG_CLIENT_ERROR",
                    "errorMessage": "Unable to find assertion on the network!",
                }
                get_public_operation_result["status"] = "FAILED"

            return {
                "operation": {
                    "get": get_operation_status_object(
                        get_public_operation_result, get_public_operation_id
                    ),
                },
            }

        if validate:
            is_valid = True  # #TODO: Implement assertion validation logic
            if not is_valid:
                get_public_operation_result["data"] = {
                    "error_type": "DKG_CLIENT_ERROR",
                    "error_message": "Calculated root hashes don't match!",
                }

        formatted_assertion = "\n".join(
            assertion.get("public", [])
            + (
                assertion.get("private")
                if isinstance(assertion.get("private"), list)
                else []
            )
        )

        formatted_metadata = None
        if output_format == OutputTypes.JSONLD.value:
            formatted_assertion = self.to_jsonld(formatted_assertion)

            if include_metadata:
                formatted_metadata = self.to_jsonld("\n".join(metadata))

        if output_format == OutputTypes.NQUADS.value:
            formatted_assertion = self.to_nquads(
                formatted_assertion, DEFAULT_RDF_FORMAT
            )
            if include_metadata:
                formatted_metadata = self.to_nquads(
                    "\n".join(metadata), DEFAULT_RDF_FORMAT
                )

        result = {
            "assertion": formatted_assertion,
            "operation": {
                "get": get_operation_status_object(
                    get_public_operation_result, get_public_operation_id
                ),
            },
        }

        if include_metadata and metadata:
            result["metadata"] = formatted_metadata

        return result

    # _extend_storing_period = Method(BlockchainRequest.extend_asset_storing_period)

    # async def extend_storing_period(
    #     self,
    #     ual: UAL,
    #     additional_epochs: int,
    #     token_amount: Wei | None = None,
    # ) -> dict[str, UAL | TxReceipt]:
    #     parsed_ual = parse_ual(ual)
    #     blockchain_id, content_asset_storage_address, token_id = (
    #         parsed_ual["blockchain"],
    #         parsed_ual["contract_address"],
    #         parsed_ual["token_id"],
    #     )

    #     if token_amount is None:
    #         latest_finalized_state = self._get_latest_assertion_id(token_id)
    #         latest_finalized_state_size = self._get_assertion_size(
    #             latest_finalized_state
    #         )

    #         token_amount = await self.node_service.get_bid_suggestion(
    #             blockchain_id,
    #             additional_epochs,
    #             latest_finalized_state_size,
    #             content_asset_storage_address,
    #             latest_finalized_state,
    #             DefaultParameters.HASH_FUNCTION_ID.value,
    #             token_amount or BidSuggestionRange.LOW,
    #         )

    #     receipt: TxReceipt = self._extend_storing_period(
    #         token_id, additional_epochs, token_amount
    #     )

    #     return {
    #         "UAL": ual,
    #         "operation": json.loads(Web3.to_json(receipt)),
    #     }

    _get_block = Method(BlockchainRequest.get_block)

    _get_assertion_size = Method(BlockchainRequest.get_assertion_size)

    def to_jsonld(self, nquads: str):
        options = {
            "algorithm": "URDNA2015",
            "format": "application/n-quads",
        }

        return jsonld.from_rdf(nquads, options)

    def to_nquads(self, content, input_format):
        options = {
            "algorithm": "URDNA2015",
            "format": "application/n-quads",
        }

        if input_format:
            options["inputFormat"] = input_format
        try:
            jsonld_data = jsonld.from_rdf(content, options)
            canonized = jsonld.to_rdf(jsonld_data, options)

            if isinstance(canonized, str):
                return [line for line in canonized.split("\n") if line.strip()]

        except Exception as e:
            raise ValueError(f"Error processing content: {e}")



from enum import Enum, auto
from dkg.types import AutoStrEnumUpperCase

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

DEFAULT_CANON_ALGORITHM = "URDNA2015"
DEFAULT_RDF_FORMAT = "application/n-quads"

PRIVATE_ASSERTION_PREDICATE = (
    "https://ontology.origintrail.io/dkg/1.0#privateMerkleRoot"
)

PRIVATE_HASH_SUBJECT_PREFIX = "https://ontology.origintrail.io/dkg/1.0#metadata-hash:"

PRIVATE_RESOURCE_PREDICATE = (
    "https://ontology.origintrail.io/dkg/1.0#representsPrivateResource"
)

CHUNK_BYTE_SIZE = 32

MAX_FILE_SIZE = 10000000


class DefaultParameters(Enum):
    ENVIRONMENT: str = "mainnet"
    PORT: int = 8900
    FREQUENCY: int = 5
    MAX_NUMBER_OF_RETRIES: int = 5
    HASH_FUNCTION_ID: int = 1
    MIN_NUMBER_OF_FINALIZATION_CONFIRMATION: int = 3
    IMMUTABLE: bool = False
    VALIDATE: bool = True
    OUTPUT_FORMAT: str = "JSON-LD"
    STATE: None = None
    INCLUDE_METADATA: bool = False
    CONTENT_TYPE: str = "all"
    HANDLE_NOT_MINED_ERROR: bool = False
    SIMULATE_TXS: bool = False
    FORCE_REPLACE_TXS: bool = False
    GAS_LIMIT_MULTIPLIER: int = 1
    PARANET_UAL: None = None
    GET_SUBJECT_UAL: bool = False
    REPOSITORY: str = "dkg"


class OutputTypes(Enum):
    NQUADS: str = "N-QUADS"
    JSONLD: str = "JSON-LD"


class Environments(Enum):
    DEVELOPMENT: str = "development"
    DEVNET: str = "devnet"
    TESTNET: str = "testnet"
    MAINNET: str = "mainnet"


class BlockchainIds(Enum):
    HARDHAT_1: str = "hardhat1:31337"
    HARDHAT_2: str = "hardhat2:31337"
    BASE_DEVNET: str = "base:84532"
    GNOSIS_DEVNET: str = "gnosis:10200"
    NEUROWEB_DEVNET: str = "otp:2160"
    BASE_TESTNET: str = "base:84532"
    GNOSIS_TESTNET: str = "gnosis:10200"
    NEUROWEB_TESTNET: str = "otp:20430"
    BASE_MAINNET: str = "base:8453"
    GNOSIS_MAINNET: str = "gnosis:100"
    NEUROWEB_MAINNET: str = "otp:2043"


class OperationStatuses(str, Enum):
    PENDING = "PENDING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"


ZERO_ADDRESS = "0x0000000000000000000000000000000000000000"
BLOCKCHAINS = {
    "development": {
        "hardhat1:31337": {
            "hub": "0x5FbDB2315678afecb367f032d93F642f64180aa3",
            "rpc": "http://localhost:8545",
        },
        "hardhat2:31337": {
            "hub": "0x5FbDB2315678afecb367f032d93F642f64180aa3",
            "rpc": "http://localhost:9545",
        },
    },
    "devnet": {
        "base:84532": {
            "hub": "0xE043daF4cC8ae2c720ef95fc82574a37a429c40A",
            "rpc": "https://sepolia.base.org",
        }
    },
    "testnet": {
        "base:84532": {
            "hub": "0xf21CE8f8b01548D97DCFb36869f1ccB0814a4e05",
            "rpc": "https://sepolia.base.org",
        },
        "gnosis:10200": {
            "hub": "0x2c08AC4B630c009F709521e56Ac385A6af70650f",
            "rpc": "https://rpc.chiadochain.net",
        },
        "otp:20430": {
            "hub": "0xd7d073b560412c6A7F33dD670d323D01061E5DEb",
            "rpc": "https://lofar-testnet.origin-trail.network",
        },
    },
    "mainnet": {
        "base:8453": {
            "hub": "0x99Aa571fD5e681c2D27ee08A7b7989DB02541d13",
            "rpc": "https://mainnet.base.org",
        },
        "gnosis:100": {
            "hub": "0x882D0BF07F956b1b94BBfe9E77F47c6fc7D4EC8f",
            "rpc": "https://rpc.gnosischain.com/",
        },
        "otp:2043": {
            "hub": "0x0957e25BD33034948abc28204ddA54b6E1142D6F",
            "rpc": "https://astrosat-parachain-rpc.origin-trail.network",
        },
    },
}

DEFAULT_PROXIMITY_SCORE_FUNCTIONS_PAIR_IDS = {
    "development": {"hardhat1:31337": 2, "hardhat2:31337": 2, "otp:2043": 2},
    "devnet": {
        "otp:2160": 2,
        "gnosis:10200": 2,
        "base:84532": 2,
    },
    "testnet": {
        "otp:20430": 2,
        "gnosis:10200": 2,
        "base:84532": 2,
    },
    "mainnet": {
        "otp:2043": 2,
        "gnosis:100": 2,
        "base:8453": 2,
    },
}

PRIVATE_HISTORICAL_REPOSITORY = "privateHistory"
PRIVATE_CURRENT_REPOSITORY = "privateCurrent"


class Operations(Enum):
    PUBLISH = "publish"
    GET = "get"
    QUERY = "query"
    PUBLISH_PARANET = "publishParanet"
    FINALITY = "finality"


class Status(AutoStrEnumUpperCase):
    ERROR = auto()
    NOT_FINALIZED = auto()
    FINALIZED = auto()
    NETWORK_ERROR = auto()


class ErrorType(AutoStrEnumUpperCase):
    DKG_CLIENT_ERROR = auto()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Any

import aiohttp
from dkg.dataclasses import HTTPRequestMethod, NodeResponseDict
from dkg.exceptions import HTTPRequestMethodNotSupported, NodeRequestError
from dkg.types import URI
from dkg.providers.node.base_node_http import BaseNodeHTTPProvider


class AsyncNodeHTTPProvider(BaseNodeHTTPProvider):
    def __init__(
        self,
        endpoint_uri: URI | str,
        api_version: str = "v1",
        auth_token: str | None = None,
    ):
        super().__init__(endpoint_uri, api_version, auth_token)

    async def make_request(
        self,
        method: HTTPRequestMethod,
        path: str,
        params: dict[str, Any] = {},
        data: dict[str, Any] = {},
    ) -> NodeResponseDict:
        url = f"{self.url}/{path}"

        async with aiohttp.ClientSession() as session:
            try:
                if method == HTTPRequestMethod.GET:
                    async with session.get(
                        url, params=params, headers=self.headers
                    ) as response:
                        response.raise_for_status()
                        response = await response.json()
                elif method == HTTPRequestMethod.POST:
                    async with session.post(
                        url, json=data, headers=self.headers
                    ) as response:
                        response.raise_for_status()
                        response = await response.json()
                else:
                    raise HTTPRequestMethodNotSupported(
                        f"{method.name} method isn't supported"
                    )

                try:
                    return NodeResponseDict(response)
                except ValueError as err:
                    raise NodeRequestError(f"JSON decoding failed: {err}")

            except aiohttp.ClientError as err:
                raise NodeRequestError(f"Network error: {err}")
            except Exception as err:
                raise NodeRequestError(f"Unexpected error: {err}")



import json
from collections import namedtuple
from pathlib import Path
from typing import Any, Type

from dkg.constants import BLOCKCHAINS
from dkg.exceptions import (
    EnvironmentNotSupported,
    RPCURINotDefined,
)
from dkg.types import URI, DataHexStr, Environment, Wei
from eth_account.signers.local import LocalAccount
from eth_typing import ABI, ABIFunction
from web3.logs import DISCARD
from web3.middleware import SignAndSendRawMiddlewareBuilder
from web3.types import TxReceipt


class BaseBlockchainProvider:
    CONTRACTS_METADATA_DIR = Path(__file__).parents[2] / "data/interfaces"

    def __init__(
        self,
        environment: Environment,
        blockchain_id: str,
        rpc_uri: URI | None = None,
        gas_price: Wei | None = None,
    ):
        if environment not in BLOCKCHAINS.keys():
            raise EnvironmentNotSupported(f"Environment {environment} isn't supported!")

        self.environment = environment
        self.rpc_uri = rpc_uri
        self.blockchain_id = (
            blockchain_id
            if blockchain_id in BLOCKCHAINS[self.environment].keys()
            else None
        )

        if self.rpc_uri is None and self.blockchain_id is not None:
            self.blockchain_id = blockchain_id
            self.rpc_uri = self.rpc_uri or BLOCKCHAINS[self.environment][
                self.blockchain_id
            ].get("rpc", None)

        if self.rpc_uri is None:
            raise RPCURINotDefined(
                "No RPC URI provided for unrecognized "
                f"blockchain ID {self.blockchain_id}"
            )

        self.gas_price = gas_price

        self.abi = self._load_abi()
        self.output_named_tuples = self._generate_output_named_tuples()

    def _generate_output_named_tuples(self) -> dict[str, dict[str, Type[tuple]]]:
        def generate_output_namedtuple(function_abi: ABIFunction) -> Type[tuple] | None:
            output_names = [output["name"] for output in function_abi["outputs"]]
            if all(name != "" for name in output_names):
                return namedtuple(f"{function_abi['name']}Result", output_names)
            return None

        output_named_tuples = {}
        for contract_name, contract_abi in self.abi.items():
            output_named_tuples[contract_name] = {}
            for item in contract_abi:
                if (item["type"] != "function") or not item["outputs"]:
                    continue
                elif item["name"] in output_named_tuples[contract_name]:
                    continue
                named_tuple = generate_output_namedtuple(item)
                if named_tuple is not None:
                    output_named_tuples[contract_name][item["name"]] = named_tuple

        return output_named_tuples

    def _load_abi(self) -> ABI:
        abi = {}

        for contract_metadata in self.CONTRACTS_METADATA_DIR.glob("*.json"):
            with open(contract_metadata, "r") as metadata_json:
                abi[contract_metadata.stem] = json.load(metadata_json)

        return abi

    def decode_logs_event(
        self, receipt: TxReceipt, contract_name: str, event_name: str
    ) -> Any:
        return (
            self.contracts[contract_name]
            .events[event_name]()
            .process_receipt(receipt, errors=DISCARD)
        )

    def set_account(self, private_key: DataHexStr):
        self.account: LocalAccount = self.w3.eth.account.from_key(private_key)
        self.w3.middleware_onion.inject(
            SignAndSendRawMiddlewareBuilder.build(private_key),
            layer=0,
        )
        self.w3.eth.default_account = self.account.address
